//////////////////////////////////////////////////////////////////////////////////
// Engineer:       James C. Ahlstrom
// 
// Create Date:    7 December 2009
// Design Name:    Digital receiver and transmitter
// Module Name:    Receiver
// Project Name:   Transceiver
// Target Devices: Cyclone 3
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////



module Receiver( clock, adc_data, eth_nRESET, rx_tune_phase, tx_tune_phase, rx_ctrl, nbRef, mem_data,
	mem_write_addr, mem_wen, mem_checksum, mem_block, rf_ref, vna_count, coefReset, coefWR, coef, decA, decA_avail, decB, decB_avail, tx_sample, txOutput);

	input clock;
	input [13:0] adc_data;
	input eth_nRESET;
	input [31:0] rx_tune_phase;
	input [31:0] tx_tune_phase;
	input [7:0] rx_ctrl;
	input [15:0] nbRef;
	output reg [31:0] mem_data;
	output reg [9:0] mem_write_addr;
	output reg mem_wen;
	output reg [31:0] mem_checksum; 		// must be unsigned
	output reg mem_block;		// Index of available block 0 or 1
	output [13:0] rf_ref;				// RF equal to Rx frequency for VNA
	input [15:0] vna_count;			// number of VNA scan frequencies, or zero
	input coefReset;
	input coefWR;
	input [35:0] coef; 
	output [31:0] decA;
	output decA_avail;
	output [31:0] decB;
	output decB_avail;
	input [31:0] tx_sample; // current tx sample for DPD
	input txOutput;
	
	`include "gparam.v"

	wire vna_avail;
	wire signed [23:0] vna_real, vna_imag;
	wire [31:0] tune_phase;
	wire decimA_avail, decimB_avail, decimC_avail;
	wire signed [17:0] decimA_real, decimA_imag;
	wire signed [17:0] decimB_real, decimB_imag;
	wire signed [17:0] decimC_real, decimC_imag;
	wire signed [23:0] decim_real, decim_imag;

	//assign decA = {decimA_imag[15:0], decimA_real[15:0]};
	assign decA = {decimA_imag[16:1] + decimA_imag[0], decimA_real[16:1] + decimA_real[0]};
	assign decA_avail = decimA_avail;
	//assign decB = {decimB_imag[15:0], decimB_real[15:0]};
	assign decB = {decimB_imag[16:1] + decimB_imag[0], decimB_real[16:1] + decimB_real[0]};
	assign decB_avail = decimB_avail;

	// The samples can come from the receiver decimators, or from the VNA module.
	wire use_vna;
	assign use_vna = |vna_count;	// use_vna is True if vna_count != 0
	wire data_avail;
	wire signed [23:0] data_real, data_imag;

	reg [5:0] decim1, decim2;
	reg [1:0] byteMode;
	reg signed [23:0] realVal0, imagVal0;
	reg signed [23:0] realVal1, imagVal1;
	wire decim_avail;
	assign data_avail = use_vna ? vna_avail : decim_avail;
	assign data_real  = use_vna ? vna_real  : decim_real;
	assign data_imag  = use_vna ? vna_imag  : decim_imag;

	// Memory blocks for ADC data
	reg [31:0] tx_data1;
	reg [31:0] tx_data2;
	reg [31:0] mem_data3;		// data to write
	reg [15:0] mem_save;		// odd two bytes to save from 6-byte sample
	reg [31:0] checksum;		// must be unsigned
	reg [3:0] rx_state;		// state machine for receive

	parameter sRx		= 0;
	parameter sRx1		= 1;
	parameter sRx2		= 2;
	parameter sRx3		= 3;
	parameter sRx4		= 4;
	parameter sRx5		= 5;
	parameter sRx6		= 6;
	parameter sRxW		= 7;
	parameter sRxB		= 8;
	parameter sRxB1		= 9;
	parameter sRxWTx1	= 10;
	parameter sRxWTx2	= 11;
	parameter sRxWTx3	= 12;
	parameter sRxWTx4	= 13;

	// Tune ADC data with CORDIC
	wire signed [17:0] cordic_real, cordic_imag;

	cordic2 rx_cordic (adc_data, clock, tune_phase, cordic_real, cordic_imag, rf_ref);
	
	// Receive CIC filters followed by FIR filter
	CicVarDecimM5 cic1 (decim1, clock, ~use_vna, cordic_real, cordic_imag, decimA_avail, decimA_real, decimA_imag);
	CicVarDecimM5 cic2 (decim2, clock, decimA_avail, decimA_real, decimA_imag, decimB_avail, decimB_real, decimB_imag);
	NoiceBlanker nb(clock, decimB_avail, decimB_real, decimB_imag, decimC_avail, decimC_real, decimC_imag, nbRef);
	firX8R8 fir2 (clock, decimC_avail, decimC_real, decimC_imag, decim_avail, decim_real, decim_imag, coefReset, coefWR, coef);

	// This the the VNA module.  It runs as an alternative to the above filters.
	VNA m_vna (clock, eth_nRESET, cordic_real, cordic_imag, vna_count,
		rx_tune_phase, tx_tune_phase, tune_phase,
		vna_avail, vna_real, vna_imag);
	
	wire [17:0] csum;	// unsigned for checksum
	assign csum = mem_data[15:0] + mem_data[31:16];

	initial
	begin
			decim1 <= 'd3;
			decim2 <= 'd3;
			byteMode <= 'd3;
			checksum <= 0;
	end
	
	always @(rx_ctrl)
	begin
		case (rx_ctrl[7:6])
			0: 
				begin
				decim1 <= 'd7;
				byteMode <= 'd3;
				end
			1: 
				begin
				decim1 <= 'd1;
				byteMode <= 'd2;
				end
			2: 
				begin
				decim1 <= 'd39;
				byteMode <= 'd3;
				end
			3: 
				begin
				decim1 <= 'd1; 
				byteMode <= 'd1;
				end
		endcase
		decim2 <= rx_ctrl[5:0];
	end

	always @(posedge clock)		// Rx state machine
	begin
		if (eth_nRESET == 0)	// we are in reset
			begin
				mem_write_addr <= 1'd0;
				mem_wen <= 0;
				case (byteMode) 
					0: rx_state <= sRx;
					1: rx_state <= sRxB;
					2: rx_state <= sRxW;
					3: rx_state <= sRx;
				endcase
			end
		else 
			case (rx_state)
			sRxB:
			begin
				if (data_avail)
				begin
					mem_save <= {data_imag[16-:8],data_real[16-:8]};
					rx_state <= sRxB1;
				end
			end
			sRxB1:
			begin
				if (data_avail)
				begin
					mem_wen <= 1'd1;
					mem_data <= {data_imag[16-:8],data_real[16-:8], mem_save};
					rx_state <= sRx5;
				end
			end
			sRxW:
			begin
				if (data_avail)
				begin
					mem_wen <= 1'd1;
					mem_data <= {data_imag[23-:16],data_real[23-:16]};
					rx_state <= sRx5;
				end
			end
			sRx:		// Wait for first sample of two samples
			begin
				if (data_avail)
				begin		// write first dword, save last two bytes LITTLE-ENDIAN
					mem_wen <= 1'd1;
					tx_data1 <= tx_sample;
					mem_data <= {data_imag[7:0], data_real};
					mem_save <= data_imag[23:8];
					rx_state <= sRx1;
				end
			end
			sRx1:
			begin
				checksum <= checksum + csum;
				mem_wen <= 1'd0;
				rx_state <= sRx2;
			end
			sRx2:		// Wait for second sample of two samples
			begin
				if (data_avail)
				begin
					mem_wen <= 1'd1;
					// memory write with low clock transition
					mem_write_addr <= mem_write_addr + 1'd1;
					tx_data2 <= tx_sample;
					mem_data <= {data_real[15:0], mem_save};
					mem_data3 <= {data_imag, data_real[23:16]};
					rx_state <= sRx3;
				end
			end
			sRx3:		// write second and third dword
			begin
				checksum <= checksum + csum;
				mem_wen <= 1'd0;
				rx_state <= sRx4;
			end
			sRx4:
			begin
				mem_wen <= 1'd1;
				mem_data <= mem_data3;
				mem_write_addr <= mem_write_addr + 1'd1;
				rx_state <= sRx5;
			end
			sRx5:
			begin
				checksum <= checksum + csum;
				mem_wen <= 1'd0;
				if (txOutput) 
					rx_state <= sRxWTx1;
				else
					rx_state <= sRx6;
			end
			sRxWTx1:
			begin
				mem_wen <= 1'd1;
				mem_data <= tx_data1;
				mem_write_addr <= mem_write_addr + 1'd1;
				rx_state <= sRxWTx2;
			end
			sRxWTx2:
			begin
				checksum <= checksum + csum;
				mem_wen <= 1'd0;
				rx_state <= sRxWTx3;
			end
			sRxWTx3:
			begin
				mem_wen <= 1'd1;
				mem_data <= tx_data2;
				mem_write_addr <= mem_write_addr + 1'd1;
				rx_state <= sRxWTx4;
			end
			sRxWTx4:
			begin
				checksum <= checksum + csum;
				mem_wen <= 1'd0;
				rx_state <= sRx6;
			end
			sRx6:		// increment memory address
			begin
				if (mem_write_addr == Transceiver.MEM_DWORD_SIZE - 1)
				begin		// first block of memory is ready to be sent
					mem_write_addr <= Transceiver.MEM_DWORD_SIZE;
					mem_checksum <= checksum;
					mem_block <= 1'd0;
					checksum <= 1'd0;
				end
				else if (mem_write_addr == Transceiver.MEM_DWORD_SIZE * 2 - 1)
				begin		// second block of memory is ready to be sent
					mem_write_addr <= 1'd0;
					mem_checksum <= checksum;
					mem_block <= 1'd1;
					checksum <= 1'd0;
				end
				else
				begin
					mem_write_addr <= mem_write_addr + 1'd1;
				end
				case (byteMode) 
					0: rx_state <= sRx;
					1: rx_state <= sRxB;
					2: rx_state <= sRxW;
					3: rx_state <= sRx;
				endcase
			end
			endcase
	end

endmodule

module NoiceBlanker(clock, x_avail, x_real, x_imag, y_avail, y_real, y_imag, refMax);
	input clock;
	input x_avail;
	input signed [17:0] x_real;
	input signed [17:0] x_imag;
	output reg y_avail;
	output reg signed [17:0] y_real;
	output reg signed [17:0] y_imag;
	input [15:0] refMax;

	reg [17:0] x_abs;
	reg signed [17:0] tmp_imag;
	reg signed [17:0] tmp_real;

	initial 
	begin
		x_abs <= 0;
	end

	always @(posedge clock)
	begin
		if (x_abs[17:2] < refMax)
		begin
			y_real <= tmp_real;
			y_imag <= tmp_imag;
		end
		else
		begin
			y_real <= 0; 
			y_imag <= 0;
		end
		y_avail <= x_avail;
		x_abs <= (x_real>0 ? x_real: -x_real) + (x_imag>0 ? x_imag : -x_imag);
		tmp_real <= x_real;
		tmp_imag <= x_imag;
	end

endmodule


module VNA (
	input clock,
	input eth_nRESET,
	input signed [17:0] cordic_real,
	input signed [17:0] cordic_imag,
	input [15:0] vna_count,
	input [31:0] rx_tune_phase,
	input [31:0] tx_tune_phase,
	output reg [31:0] tune_phase,
	output reg vna_avail,
	output reg signed [23:0] vna_real,
	output reg signed [23:0] vna_imag
	);

	parameter CountWait = 14'd8000;
	parameter CountData = 14'd4096;
	reg [13:0] counter;
	reg [15:0] npixels;
	reg signed [29:0] sum_real, sum_imag;

	// State machine for VNA
	reg [3:0] vna_state;
	parameter sVna0		= 0;
	parameter sVna1		= 1;
	parameter sVna2		= 2;
	parameter sVna3		= 3;
	parameter sVna4		= 4;
	parameter sVna5		= 5;
	parameter sVnaReset	= 6;

	initial
	begin
		tune_phase <= 0;
		vna_avail <= 0;
		vna_real <= 0;
		vna_imag <= 0;
	end

	always @(posedge clock)		// VNA state machine
	begin
		if (eth_nRESET == 0 || vna_count == 0)	// we are in reset, or non-VNA
		begin
			tune_phase <= rx_tune_phase;
			vna_state <= sVnaReset;
		end
		else case (vna_state)
		sVnaReset:
		begin
			npixels <= vna_count;
			counter <= CountWait;
			vna_state <= sVna0;
		end
		sVna0:		// Throw away data after frequency change
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
			end
			else
			begin
				vna_state <= sVna1;
				counter <= CountData;
			end
		end
		sVna1:		// Sum data
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
				sum_real <= sum_real + cordic_real;
				sum_imag <= sum_imag + cordic_imag;
			end
			else
			begin
				vna_state <= sVna2;
				counter <= CountWait;
			end
		end
		sVna2:		// Send an output datum
		begin
			vna_avail <= 1;
			vna_real <= sum_real[29:6];
			vna_imag <= sum_imag[29:6];
			vna_state <= sVna3;
		end
		sVna3:		// Increment the frequency, and repeat
		begin
			vna_avail <= 0;
			sum_real <= 1'd0;
			sum_imag <= 1'd0;
			if (npixels != 1'd1)
			begin
				npixels <= npixels - 1'd1;
				tune_phase <= tune_phase + tx_tune_phase;
				vna_state <= sVna0;
			end
			else
			begin
				npixels <= vna_count;
				tune_phase <= rx_tune_phase;
				vna_state <= sVna4;
			end
		end
		sVna4:		// Frequency reset to start, send a zero sample to mark the start
		begin
			if (counter)
			begin
				counter <= counter - 1'd1;
			end
			else
			begin
				vna_avail <= 1'd1;
				vna_real <= 1'd0;
				vna_imag <= 1'd0;
				vna_state <= sVna5;
			end
		end
		sVna5:
		begin
			vna_avail <= 1'd0;
			counter <= CountWait;
			vna_state <= sVna0;
		end
		endcase
	end
endmodule
